-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("achievementTracker", "AceEvent-3.0", "AceHook-3.0")
local _, _, _, interfaceVersion = GetBuildInfo()

-----------------------
--- Library loading ---
-----------------------
local LATr = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-AchievementTracker")
local LSM = LibStub("LibSharedMedia-3.0")
local LibQTip = LibStub("LibQTip-1.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local CreateFont, CreateFrame = CreateFont, CreateFrame
local type, unpack, format, pairs = type, unpack, string.format, pairs
local GetAchievementInfo, GetAchievementLink, RemoveTrackedAchievement = GetAchievementInfo, GetAchievementLink, RemoveTrackedAchievement
local GetAchievementCategory, GetCategoryInfo, GetCategoryNumAchievements = GetAchievementCategory, GetCategoryInfo, GetCategoryNumAchievements
local GetAchievementNumCriteria, GetAchievementCriteriaInfo = GetAchievementNumCriteria, GetAchievementCriteriaInfo
local UnitAffectingCombat = UnitAffectingCombat

-----------------
--- Variables ---
-----------------
module.modName = LATr["Achievement Tracker"]
local prdb, chdb, options
local numTrackedAchievement = 0
local category, achievements = {}, {}
local achievementTracker, trackerAnchor
local outlines = {[""] = LATr["None"], ["OUTLINE"] = LATr["Outline"], ["THICKOUTLINE"] = LATr["Thick Outline"]}
local cATrFont = CreateFont("cATrFont")
local scATrFont = CreateFont("scATrFont")
local aATrFont = CreateFont("aATrFont")
local acATrFont = CreateFont("acATrFont")

----------------------------
--- Tracker Anchor Frame ---
----------------------------
local anchor = CreateFrame("Button", nil, UIParent)
anchor:SetHeight(24)
anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
local text = anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
text:SetPoint("CENTER")
text:SetText(LATr["Achievement Tracker"])
anchor:SetWidth(text:GetStringWidth() + 8)
anchor:SetMovable(true)
anchor:RegisterForDrag("LeftButton")
anchor:SetClampedToScreen(true)
anchor:SetScript("OnDragStart", function(self)
	self:StartMoving()
end)
anchor:SetScript("OnDragStop", function(self)
	self:StopMovingOrSizing()
	prdb.point, prdb.x, prdb.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	LibQTip:Release(achievementTracker)
	module:updateAchievementTracker()
end)

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		point = "CENTER",
		x = 0,
		y = 0,
		showAnchor = true,
		anchorPosition = 2,
		toggleTracker = true,
		hideTrackerInCombat = true,
		categoryFont = LSM.DefaultMedia.font,
		categoryFontSize = 12,
		categoryFontOutline = "",
		categoryCollapseButtonSize = 24,
		subCategoryFont = LSM.DefaultMedia.font,
		subCategoryFontSize = 12,
		subCategoryFontOutline = "",
		subCategoryCollapseButtonSize = 24,
		achievementFont = LSM.DefaultMedia.font,
		achievementFontSize = 10,
		achievementFontOutline = "",
		achievementImageSize = 24,
		achievementCriteriaFont = LSM.DefaultMedia.font,
		achievementCriteriaFontSize = 8,
		achievementCriteriaFontOutline = "",
		maxWidth = 300,
		subCategoryCellPadding = 5,
		achievementCellPadding = 10,
		criteriasCellPadding = 15,
		scale = 1,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
	},
}

---------------
--- Options ---
---------------
local optGetter, optSetter
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		LibQTip:Release(achievementTracker)
		module:updateAchievementTracker()
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			desc = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				showAnchor = {
					name = LATr["Show achievement tracker anchor"],
					type = "toggle",
					set = function(info, value)
						prdb[info[#info]] = value
						if value then
							anchor:Show()
						else
							anchor:Hide()
						end
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 6,
				},
				anchorPosition = {
					name = LATr["Anchor position"],
					type = "select",
					values = {[1] = "TOPLEFT", [2] = "TOP", [3] = "TOPRIGHT", [4] = "BOTTOMLEFT", [5] = "BOTTOM", [6] = "BOTTOMRIGHT"},
					width = "full",
					order = 7,
				},
				toggleTracker = {
					name = LATr["Display tracker"],
					type = "toggle",
					width = "full",
					order = 9,
				},
				hideTrackerInCombat = {
					name = LATr["Hide achievement tracker in combat"],
					type = "toggle",
					set = function(info, value)
						prdb[info[#info]] = value
						if value then
							module:RegisterEvent("PLAYER_REGEN_ENABLED","updateAchievementTracker")
							module:RegisterEvent("PLAYER_REGEN_DISABLED","updateAchievementTracker")
						else
							module:UnregisterEvent("PLAYER_REGEN_ENABLED")
							module:UnregisterEvent("PLAYER_REGEN_DISABLED")
						end
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 10,
				},
				categoryFont = {
					name = LATr["Achievement category font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = function(info, value)
						prdb[info[#info]] = value
						cATrFont:SetFont(LSM:Fetch("font", value), prdb.categoryFontSize, prdb.categoryFontOutline)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 11,
				},
				categoryFontSize = {
					name = LATr["Achievement category font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = function(info, value)
						prdb[info[#info]] = value
						cATrFont:SetFont(LSM:Fetch("font", prdb.categoryFont), value, prdb.categoryFontOutline)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 12,
				},
				categoryFontOutline = {
					name = LATr["Achievement category font outline"],
					type = "select",
					values = outlines,
					set = function(info, value)
						prdb[info[#info]] = value
						cATrFont:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, value)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
						end,
					width = "full",
					order = 13,
				},
				categoryCollapseButtonSize = {
					name = LATr["Category collapse Button size"],
					type = "range",
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 14,
				},
				subCategoryFont = {
					name = LATr["Achievement sub category font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = function(info, value)
						prdb[info[#info]] = value
						scATrFont:SetFont(LSM:Fetch("font", value), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 15,
				},
				subCategoryFontSize = {
					name = LATr["Achievement sub category font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = function(info, value)
						prdb[info[#info]] = value
						scATrFont:SetFont(LSM:Fetch("font", prdb.subCategoryFont), value, prdb.subCategoryFontOutline)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 16,
				},
				subCategoryFontOutline = {
					name = LATr["Achievement sub category font outline"],
					type = "select",
					values = outlines,
					set = function(info, value)
						prdb[info[#info]] = value
						scATrFont:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, value)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
						end,
					width = "full",
					order = 17,
				},
				subCategoryCollapseButtonSize = {
					name = LATr["Sub category collapse Button size"],
					type = "range",
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 18,
				},
				achievementFont = {
					name = LATr["Achievement font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = function(info, value)
						prdb[info[#info]] = value
						aATrFont:SetFont(LSM:Fetch("font", value), prdb.achievementFontSize, prdb.achievementFontOutline)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 19,
				},
				achievementFontSize = {
					name = LATr["Achievement font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = function(info, value)
						prdb[info[#info]] = value
						aATrFont:SetFont(LSM:Fetch("font", prdb.achievementFont), value, prdb.achievementFontOutline)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 20,
				},
				achievementFontOutline = {
					name = LATr["Achievement font outline"],
					type = "select",
					values = outlines,
					set = function(info, value)
						prdb[info[#info]] = value
						aATrFont:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, value)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
						end,
					width = "full",
					order = 21,
				},
				achievementImageSize = {
					name = LATr["Achievement image size"],
					type = "range",
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 22,
				},
				achievementCriteriaFont = {
					name = LATr["Achievement criterias font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = function(info, value)
						prdb[info[#info]] = value
						acATrFont:SetFont(LSM:Fetch("font", value), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 23,
				},
				achievementCriteriaFontSize = { 
					name = LATr["Achievement criterias font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = function(info, value)
						prdb[info[#info]] = value
						acATrFont:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), value, prdb.achievementCriteriaFontOutline)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 24,
				},
				achievementCriteriaFontOutline = {
					name = LATr["Achievement criterias font outline"],
					type = "select",
					values = outlines,
					set = function(info, value)
						prdb[info[#info]] = value
						acATrFont:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, value)
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
						end,
					width = "full",
					order = 25,
				},
				maxWidth = {
					name = LATr["Achievement tracker column max width"],
					type="range",
					min = 50,
					max = 600,
					step = 5,
					width = "full",
					order = 26,
				},
				subCategoryCellPadding = {
					name = LATr["Sub category cell padding"],
					type="range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 27,
				},
				achievementCellPadding = {
					name = LATr["Achievement cell padding"],
					type="range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 28,
				},
				criteriasCellPadding = {
					name = LATr["Achievement criterias cell padding"],
					type="range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 29,
				},
				scale = {
					name = LATr["Achievement tracker scale"],
					type = "range",
					min = 0,
					max = 1,
					step = 0.1,
					width = "full",
					order = 30,
				},
				backDropColor = {
					name = LATr["Achievement tracker background color"],
					type = "color",
					hasAlpha = true,
					get = function(info)
						return unpack(prdb[info[#info]])
					end,
					set = function(info, arg1, arg2, arg3, arg4)
						prdb[info[#info]][1] = arg1
						prdb[info[#info]][2] = arg2
						prdb[info[#info]][3] = arg3
						prdb[info[#info]][4] = arg4
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 31,
				},
				borderColor = {
					name = LATr["Achievement tracker border color"],
					type = "color",
					hasAlpha = true,
					get = function(info)
						return unpack(prdb[info[#info]])
					end,
					set = function(info, arg1, arg2, arg3, arg4)
						prdb[info[#info]][1] = arg1
						prdb[info[#info]][2] = arg2
						prdb[info[#info]][3] = arg3
						prdb[info[#info]][4] = arg4
						LibQTip:Release(achievementTracker)
						module:updateAchievementTracker()
					end,
					width = "full",
					order = 32,
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("achievementTracker", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	self:SetEnabledState(core:GetModuleEnabled("achievementTracker"))
	core:RegisterModuleOptions("achievementTracker", getOptions, module.modName)
	chdb.achievementTracked = chdb.achievementTracked or {}
	chdb.categoryDisplay = chdb.categoryDisplay or {}
	if prdb.hideTrackerInCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "updateAchievementTracker")
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "updateAchievementTracker")
	end
	if interfaceVersion == 30100 then
		self:SecureHook("AddTrackedAchievement")
		self:SecureHook("RemoveTrackedAchievement")
		self:SecureHook("GetNumTrackedAchievements")
		self:RegisterEvent("TRACKED_ACHIEVEMENT_UPDATE")
	end
end

function module:OnEnable()
	cATrFont:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
	scATrFont:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
	aATrFont:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achivementFontOutline)
	acATrFont:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
end

function module:OnDisable()
	anchor:Hide()
	LibQTip:Release(achievementTracker)
	achievementTracker = nil
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
	chdb = self.db.char
	if not self:IsEnabled() then return end
	cATrFont:SetFont(LSM:Fetch("font", prdb.categoryFont), prdb.categoryFontSize, prdb.categoryFontOutline)
	scATrFont:SetFont(LSM:Fetch("font", prdb.subCategoryFont), prdb.subCategoryFontSize, prdb.subCategoryFontOutline)
	aATrFont:SetFont(LSM:Fetch("font", prdb.achievementFont), prdb.achievementFontSize, prdb.achivementFontOutline)
	acATrFont:SetFont(LSM:Fetch("font", prdb.achievementCriteriaFont), prdb.achievementCriteriaFontSize, prdb.achievementCriteriaFontOutline)
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
end

---------------------------
--- Main Module Program ---
---------------------------
--------------
--- Events ---
--------------
function module:TRACKED_ACHIEVEMENT_UPDATE(event, aid)
	if achievements[aid] and achievements[aid].numCriterias > 0 then
		for cid = 1, achievements[aid].numCriterias do
			local cString, cType, completed, quantity, reqQuantity, charName, flags, assetID, quantityString, cid2 = GetAchievementCriteriaInfo(aid, cid)
			achievements[aid].criterias[cid] = {}
			achievements[aid].criterias[cid].name = cString
			achievements[aid].criterias[cid].cType = cType
			achievements[aid].criterias[cid].completed = completed or 0
			achievements[aid].criterias[cid].quantity = quantity
			achievements[aid].criterias[cid].reqQuantity = reqQuantity
		end
	end
end

-------------
--- Hooks ---
-------------
function module:AddTrackedAchievement(aid)
	numTrackedAchievement = numTrackedAchievement + 1
	chdb.achievementTracked[aid] = true
	achievements[aid] = {}
	_, achievements[aid].name, _, achievements[aid].completed, _, _, _, achievements[aid].description, _, achievements[aid].image = GetAchievementInfo(aid)
	achievements[aid].numCriterias = GetAchievementNumCriteria(aid)
	achievements[aid].criterias = {}
	for cid = 1, achievements[aid].numCriterias do
		local cString, cType, completed, quantity, reqQuantity, charName, flags, assetID, quantityString, cid2 = GetAchievementCriteriaInfo(aid, cid)
		achievements[aid].criterias[cid] = {}
		achievements[aid].criterias[cid].name = cString
		achievements[aid].criterias[cid].cType = cType
		achievements[aid].criterias[cid].completed = completed or 0
		achievements[aid].criterias[cid].quantity = quantity
		achievements[aid].criterias[cid].reqQuantity = reqQuantity
	end
	local categoryName, subCategoryName
	achievements[aid].subCategory = GetAchievementCategory(aid)
	subCategoryName, achievements[aid].category = GetCategoryInfo(achievements[aid].subCategory)
	categoryName = GetCategoryInfo(achievements[aid].category)
	if achievements[aid].category and achievements[aid].category > 0 then
		if not category[achievements[aid].category] then
			category[achievements[aid].category] = {}
			category[achievements[aid].category].subCategory = {}
			category[achievements[aid].category].numSubCategory = 0
			category[achievements[aid].category].achievements = {}
			category[achievements[aid].category].numAchievements = 0
			category[achievements[aid].category].name = categoryName
			category[achievements[aid].category].numTA, category[achievements[aid].category].numTCA  = GetCategoryNumAchievements(achievements[aid].category)
		end
		if not category[achievements[aid].category].subCategory[achievements[aid].subCategory] then
			category[achievements[aid].category].subCategory[achievements[aid].subCategory] = {}
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements = 0
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].achievements = {}
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].name = subCategoryName
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].numTA, category[achievements[aid].category].subCategory[achievements[aid].subCategory].numTCA  = GetCategoryNumAchievements(achievements[aid].subCategory)

		end
		category[achievements[aid].category].numSubCategory = category[achievements[aid].category].numSubCategory + 1
		category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements = category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements + 1
		category[achievements[aid].category].subCategory[achievements[aid].subCategory].achievements[aid] = true
	elseif achievements[aid].subCategory and achievements[aid].subCategory > 0 then
		if not category[achievements[aid].subCategory] then
			category[achievements[aid].subCategory] = {}
			category[achievements[aid].subCategory].subCategory = {}
			category[achievements[aid].subCategory].numSubCategory = 0
			category[achievements[aid].subCategory].achievements = {}
			category[achievements[aid].subCategory].numAchievements = 0
			category[achievements[aid].subCategory].name = subCategoryName
			category[achievements[aid].subCategory].numTA, category[achievements[aid].subCategory].numTCA  = GetCategoryNumAchievements(achievements[aid].subCategory)
		end
		category[achievements[aid].subCategory].numAchievements = category[achievements[aid].subCategory].numAchievements + 1
		category[achievements[aid].subCategory].achievements[aid] = true
	end
	LibQTip:Release(achievementTracker)
	achievementTracker = nil
	module:updateAchievementTracker()
end

function module:RemoveTrackedAchievement(aid)
	numTrackedAchievement = numTrackedAchievement - 1
	chdb.achievementTracked[aid] = nil
	achievements[aid].name = nil
	achievements[aid].description = nil
	achievements[aid].image = nil
	if achievements[aid].category and achievements[aid].category > 0 then
		if achievements[aid].subCategory and achievements[aid].subCategory > 0 then
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].achievements[aid] = nil
			category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements = category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements - 1
			if category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements == 0 then
				category[achievements[aid].category].subCategory[achievements[aid].subCategory].name = nil
				category[achievements[aid].category].subCategory[achievements[aid].subCategory].numAchievements = nil
				category[achievements[aid].category].subCategory[achievements[aid].subCategory].numTA = nil
				category[achievements[aid].category].subCategory[achievements[aid].subCategory].numTCA = nil
				category[achievements[aid].category].subCategory[achievements[aid].subCategory].achievements = nil
				category[achievements[aid].category].subCategory[achievements[aid].subCategory] = nil
				category[achievements[aid].category].numSubCategory = category[achievements[aid].category].numSubCategory - 1
				if category[achievements[aid].category].numSubCategory == 0 and category[achievements[aid].category].numAchievements == 0 then
					category[achievements[aid].category].subCategory = nil
					category[achievements[aid].category].numSubCategory = nil
					category[achievements[aid].category].achievements = nil
					category[achievements[aid].category].name = nil
					category[achievements[aid].category].numAchievements = nil
					category[achievements[aid].category].numTA = nil
					category[achievements[aid].category].numTCA = nil
				end
			end
		end
	elseif achievements[aid].subCategory and achievements[aid].subCategory > 0 then
		category[achievements[aid].subCategory].achievements[aid] = nil
		category[achievements[aid].subCategory].numAchievements = category[achievements[aid].subCategory].numAchievements - 1
		if category[achievements[aid].subCategory].numSubCategory == 0 and category[achievements[aid].subCategory].numAchievements == 0 then
			category[achievements[aid].subCategory].subCategory = nil
			category[achievements[aid].subCategory].numSubCategory = nil
			category[achievements[aid].subCategory].achievements = nil
			category[achievements[aid].subCategory].name = nil
			category[achievements[aid].subCategory].numAchievements = nil
			category[achievements[aid].subCategory].numTA = nil
			category[achievements[aid].subCategory].numTCA = nil
		end
	end
	LibQTip:Release(achievementTracker)
	achievementTracker = nil
	module:updateAchievementTracker()
end

function module:GetNumTrackedAchievements()
	return numTrackedAchievement
end

----------------------------
--- Custom Cell Provider ---
----------------------------
local cProvider, cPrototype, basecCellPrototype = LibQTip:CreateCellProvider(LibQTip.LabelProvider)
function cPrototype:InitializeCell()
	self.fontString = self:CreateFontString()
	self.fontString:SetFontObject(GameTooltipText)
end

function cPrototype:SetupCell(tooltip, value, justification, font, bType, bValue, maxWidth, indent, anchor)
	self:EnableMouse(true)
	self:SetScript("OnMouseDown", function(frame, button)
		if bType == "category" then
			if chdb.categoryDisplay[bValue] == 1 then
				chdb.categoryDisplay[bValue] = 0
			else
				chdb.categoryDisplay[bValue] = 1
			end
			achievementTracker:SetScript("OnLeave", nil)
			LibQTip:Release(achievementTracker)
			achievementTracker = nil
			module:updateAchievementTracker()
		elseif bType == "achievement" then
			if button == "RightButton" then
				if IsShiftKeyDown() then
					ChatFrameEditBox:Show()
					ChatEdit_InsertLink(GetAchievementLink(bValue))
				else
					ItemRefTooltip:ClearLines()
					ShowUIPanel(ItemRefTooltip)
					if not ItemRefTooltip:IsVisible() then
						ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
					end
					local link = GetAchievementLink(bValue)
					if link then
						ItemRefTooltip:SetHyperlink(link)
					end
				end
			elseif button == "LeftButton" then
				if IsShiftKeyDown() then
					RemoveTrackedAchievement(bValue)
				else
					if AchievementFrame:IsShown() then
						HideUIPanel(AchievementFrame);
					else
						ShowUIPanel(AchievementFrame);
					end
				end
			end
		end
	end)
	self:SetScript("OnEnter", function()
		if bType == "achievement" then
			GameTooltip:SetOwner(tooltip, anchor)
			GameTooltip:SetHyperlink(GetAchievementLink(bValue))
		end
	end)
	self:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)

	indent = indent or 0
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font)
	local fs = self.fontString
	fs:SetPoint("TOPLEFT", self, "TOPLEFT", indent, 0)
	fs:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
	width = width + indent
	if width > maxWidth then
		fs:SetWidth(maxWidth)
		height = fs:GetHeight()
		width = maxWidth
	end
	fs:Show()
	return width, height
end

function cPrototype:ReleaseCell() 
	self:EnableMouse(false) 
	self:ClearAllPoints()
	self:SetScript("OnEnter", nil) 
	self:SetScript("OnLeave", nil) 
	self:SetScript("OnMouseDown", nil) 
end

local function getColorFromCompletionHEX(percent)
	if percent <= 0 then
		return "ff0000"
	elseif percent >= 1 then
		return "00ff00"
	elseif percent == 0.5 then
		return "ffff00"
	elseif percent < 0.5 then
		percent = percent / 0.5
		return format("%02x%02x%02x",255, percent*255, 0)
	elseif percent < 1 then
		percent = (percent-0.5) / 0.5
		return format("%02x%02x%02x",255 - 255*percent, 255, 0)
	end
end

function module:updateAchievementTracker()
	if (not prdb.toggleTracker) or (numTrackedAchievement == 0) or (prdb.hideTrackerInCombat and UnitAffectingCombat("player")) then return end
	achievementTracker = LibQTip:Acquire("tomQuest2Achievements", 2, "LEFT", "RIGHT")
	achievementTracker:EnableMouse(true)
	achievementTracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
	achievementTracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
	achievementTracker:SetScale(prdb.scale)
	achievementTracker:SetFrameStrata("BACKGROUND")

	local tempLine
	for cid, cData in pairs(category) do
		chdb.categoryDisplay[cid] = chdb.categoryDisplay[cid] or 1
		tempLine = achievementTracker:AddLine()
 		achievementTracker:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r", chdb.categoryDisplay[cid] == 0 and "Plus" or "Minus", prdb.categoryCollapseButtonSize, getColorFromCompletionHEX(cData.numTCA / cData.numTA) or "ffffff", cData.name), cATrFont, "LEFT", 1, cProvider, "category", cid, prdb.maxWidth, 0)
 		achievementTracker:SetCell(tempLine, 2, format("|cff%s%s/%s|r", getColorFromCompletionHEX(cData.numTCA / cData.numTA), cData.numTCA , cData.numTA), cATrFont, "RIGHT", 1, cProvider, "category", cid, prdb.maxWidth, 0)
		if chdb.categoryDisplay[cid] == 1 then
			if cData.numAchievements > 0 then
				for aid in pairs(cData.achievements) do
					tempLine = achievementTracker:AddLine()
 					achievementTracker:SetCell(tempLine, 1, format("|T%s:%d|t |cffffffff%s|r", achievements[aid].image, prdb.achievementImageSize, achievements[aid].name), aATrFont, "LEFT", 2, cProvider, "achievement", aid, prdb.maxWidth, prdb.subCategoryCellPadding, trackerAnchor)
				end
			end
			if category[cid].numSubCategory > 0 then
				for scid, scData in pairs(cData.subCategory) do
					chdb.categoryDisplay[scid] = chdb.categoryDisplay[scid] or 1
					tempLine = achievementTracker:AddLine()
 					achievementTracker:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r", chdb.categoryDisplay[scid] == 0 and "Plus" or "Minus", prdb.subCategoryCollapseButtonSize, getColorFromCompletionHEX(scData.numTCA / scData.numTA), scData.name), scATrFont, "LEFT", 1, cProvider, "category", scid, prdb.maxWidth, prdb.subCategoryCellPadding)
 					achievementTracker:SetCell(tempLine, 2, format("|cff%s%s/%s|r", getColorFromCompletionHEX(scData.numTCA / scData.numTA), scData.numTCA , scData.numTA), scATrFont, "RIGHT", 1, cProvider, "category", scid, prdb.maxWidth, prdb.subCategoryCellPadding)
					if chdb.categoryDisplay[scid] == 1 then
						if scData.numAchievements > 0 then
							for aid in pairs(scData.achievements) do
								tempLine = achievementTracker:AddLine()
 								achievementTracker:SetCell(tempLine, 1, format("|T%s:%d|t |cffffffff%s|r", achievements[aid].image, prdb.achievementImageSize, achievements[aid].name), aATrFont, "LEFT", 2, cProvider, "achievement", aid, prdb.maxWidth, prdb.achievementCellPadding, trackerAnchor)
							end
						end
					end
				end
			end
		end
	end
	
	achievementTracker:SetClampedToScreen(true)
	achievementTracker:ClearAllPoints()
	if prdb.anchorPosition == 1 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT")
		else
			achievementTracker:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT", 0, 24)
		end
	elseif prdb.anchorPosition == 2 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("TOP", anchor, "BOTTOM")
		else
			achievementTracker:SetPoint("TOP", anchor, "BOTTOM", 0, 24)
		end
	elseif prdb.anchorPosition == 3 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("TOPRIGHT", anchor, "BOTTOMRIGHT")
		else
			achievementTracker:SetPoint("TOPRIGHT", anchor, "BOTTOMRIGHT", 0, 24)
		end
	elseif prdb.anchorPosition == 4 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT")
		else
			achievementTracker:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 0, -24)
		end
	elseif prdb.anchorPosition == 5 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("BOTTOM", anchor, "TOP")
		else
			achievementTracker:SetPoint("BOTTOM", anchor, "TOP", 0, -24)
		end
	elseif prdb.anchorPosition == 6 then
		if prdb.showAnchor then
			achievementTracker:SetPoint("BOTTOMRIGHT", anchor, "TOPRIGHT")
		else
			achievementTracker:SetPoint("BOTTOMRIGHT", anchor, "TOPRIGHT", 0, -24)
		end
	end
	achievementTracker:Show()
end

--------------------------
--- Keybinding globals ---
--------------------------
BINDING_NAME_TQ2ACHIEVEMENTTRACKERTOGGLE = LATr["Achievement Tracker"]

function tQ2toggleAchievementTracker()
	prdb.toggleTracker = not prdb.toggleTracker
	LibQTip:Release(achievementTracker)
	module:updateAchievementTracker()
end